@@ -20,6 +20,7 @@ attr named => sub { {} };
 attr param => sub { [] };
 attr to => undef;
 attr dest => undef;
+attr psgi_middleware => undef;
 
 # helpers for matching different types of wildcards
 sub __noslash
@@ -152,7 +153,9 @@ sub _build_regex
         # - /test/ matches
         # - /test/something matches
         # - /testsomething does not match
-        $pattern .= '(?:/|$)';
+        # if the bridge is already followed by a trailing slash, it's not a
+        # concern
+        $pattern .= '(?:/|$)' unless $trailing_slash;
     }
     else {
 
@@ -405,6 +408,11 @@ The loaded destination. An array reference with two values, a controller name
 (or undef if not a controller) and the code reference to the method. It will be
 automatically generated by the router based on the contents of L</to>.
 
+=head2 psgi_middleware
+
+Extra middleware for Kelp, for this route only. It must be a code reference,
+and the middleware must wrap L<Kelp/NEXT_APP>.
+
 =head1 METHODS
 
 =head2 match
